/* splitcap.c
 * Split capture files into separate files by various criteria
 * Copyright 2003, Graeme Hewson <graeme.hewson@oracle.com>
 *
 * $Id$
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <glib.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <wiretap/wtap.h>
#include <epan/packet.h>
#include <epan/bitswap.h>
#include "ipproto.h"
#include "packet-eth.h"
#include "packet-tr.h"
#include "llcsaps.h"
#include "oui.h"
#include "packet-ip.h"
#include "packet-ipv6.h"
#include "etypes.h"

#ifdef NEED_GETOPT_H
#include "getopt.h"
#endif

/* XXX Below was nicked from packet-smb.c.  Should go into a common header. */
#ifndef CHAR_BIT
#define CHAR_BIT 8
#endif

#ifndef TIME_T_MIN
#define TIME_T_MIN ((time_t)0 < (time_t) -1 ? (time_t) 0 \
		    : ~ (time_t) 0 << (sizeof (time_t) * CHAR_BIT - 1)) 
#endif
#ifndef TIME_T_MAX
#define TIME_T_MAX (~ (time_t) 0 - TIME_T_MIN)
#endif
/* XXX Above was nicked from packet-smb.c.  Should go into a common header. */

/*---------------------------------- sll.h -----------------------------------
 * Below was taken from sll.h in libpcap or tcpdump, except that u_int8_t and
 * u_int16_t were changed to the corresponding GLib types.  See original for
 * more commentary; see also pcap(3).
 * Values for sll_hatype are defined in /usr/include/linux/if_arp.h.
 */

/*
 * A DLT_LINUX_SLL fake link-layer header
 */
#define SLL_HDR_LEN	16		/* total header length */
#define SLL_ADDRLEN	8		/* length of address field */

struct sll_header {
	guint16 sll_pkttype;		/* packet type */
	guint16 sll_hatype;			/* link-layer address type */
	guint16 sll_halen;			/* link-layer address length */
	guint8 sll_addr[SLL_ADDRLEN];	/* link-layer address */
	guint16 sll_protocol;		/* protocol */
};

/*
 * The LINUX_SLL_ values for "sll_pkttype"
 */
#define LINUX_SLL_HOST		0
#define LINUX_SLL_BROADCAST	1
#define LINUX_SLL_MULTICAST	2
#define LINUX_SLL_OTHERHOST	3
#define LINUX_SLL_OUTGOING	4

/*
 * The LINUX_SLL_ values for "sll_protocol"
 */
#define LINUX_SLL_P_802_3	0x0001	/* Novell 802.3 frames without 802.2 LLC header */
#define LINUX_SLL_P_802_2	0x0004	/* 802.2 frames (not D/I/X Ethernet) */
/*---------------------------------- sll.h -----------------------------------*/

typedef struct {
	guint16 tci;
	guint16 type;
} ether_vlan;

typedef struct {
	guint8 fc;
	guint8 dst[6];
	guint8 src[6];
} fddi_hdr;

typedef struct {
	guint8 dsap;
	guint8 ssap;
	guint8 control;
} llc;

typedef struct {
	guint8 oui[3];
	guint8 type[2];
} snap;

enum l3type {
	l3type_notinterested,
	l3type_IP_v4,
	l3type_IP_v6
};

typedef struct {
	gchar       *key;
	gchar       *filename_suffix;
	GList       *open_queue_entry;
	wtap_dumper *wr_hdl;
	int          filetype;
} dumpnode;

typedef struct {
	gchar              src_addr[INET6_ADDRSTRLEN];
	gchar              dst_addr[INET6_ADDRSTRLEN]; 
	guint8             flags;
#define IP_PAIR_SRC_GT 0x01
#define IP_PAIR_EQ     0x02
} ip_pair;

static struct split_data {
	jmp_buf            jmpenv;
    gchar             *out_dir;
    gchar             *out_fname_prefix;
	long               time_intvl;
	wtap              *rd_hdl;
	const struct wtap_pkthdr *pkthdr;
	union wtap_pseudo_header *pseudo_hdr;
	const guint8      *buf_ptr;
	const gchar       *in_fname;
	int                in_encap;
	int                snaplen;
	int                out_file_type;
	dumpnode           dflt_dumpnode;
	dumpnode          *dflt_dumpnp;
	GHashTable        *out_hash;
	GMemChunk         *dumpnode_chunk;
	GList             *out_open_queue_head;
	GList             *out_open_queue_tail;

	guint16            split_flags;
#define BI_DIR         0x01
#define OPEN_MAX_HIT   0x02
#define TIME_HH        0x04
#define TIME_SS        0x08

	guint16            split_type;
#define BY_ENCAP       0x01
#define BY_VLAN        0x02
#define BY_MAC         0x04
#define BY_IP_ADDR     0x08
#define BY_PORT        0x10
#define BY_TIME        0x20
#define NEED_IP_PAIR   (BY_IP_ADDR | BY_PORT)
} sd;

/* 1536 = 0x600 */
#define ETHERTYPE_MIN 1536
#define ETHER_STRLEN (sizeof "ff:ff:ff:ff:ff:ff" - 1)
#define FIELD_SEP '_'
#define ADDR_SEP '-'
#define PORT_SEP '#'
#define DFLT_PREFIX "trace"
#define _ABORT longjmp(sd.jmpenv, 1)

extern char  *optarg;
extern int    optind;

static void usage();
static void split(const char *const in_fname);

static void split_eth(const eth_hdr *const eth_hdrp,
	const llc **const llcp, gconstpointer *const l3hdrp,
	enum l3type *const l3type, gint32 *const framelenp,
	gchar **const ofnamepp, gboolean sep_needed);

static void split_sll(const struct sll_header *const sll_hdrp,
	const llc **const llcp, gconstpointer *const l3hdrp,
	enum l3type *const l3type, gint32 *const framelenp,
	gchar **const ofnamepp, gboolean sep_needed);

static void split_tr(const tr_hdr *const tr_hdrp, const llc **const llcp,
	gint32 *const framelenp,
	gchar **const ofnamepp, gboolean sep_needed);

static void split_fddi(const fddi_hdr *const fddi_hdrp, const llc * *const llcp,
	gint32 *const framelenp,
	gchar **const ofnamepp, gboolean sep_needed, gboolean bit_swapped);

static void split_llc(const llc *const llcp,
	gconstpointer *const l3hdrp, enum l3type *const l3type,
	gint32 *const framelenp, gchar **const ofnamepp,
	gboolean sep_needed);

static enum l3type decode_ether_type(guint16 eth_type);

static void split_ip_v4(const e_ip *const ip_hdrp, gconstpointer *const l4hdrp,
	guint8 *l4type, gint32 *const framelenp, ip_pair *const ip_pairp,
	gchar **const ofnamepp, gboolean sep_needed);

static void split_ip_v6(const struct ip6_hdr *const ip_hdrp,
	gconstpointer *const l4hdrp,
	guint8 *l4type, gint32 *const framelenp, ip_pair *const ip_pairp,
	gchar **const ofnamepp, gboolean sep_needed);

static void split_tcp(const struct tcphdr *const tcp_hdrp,
	gint32 *const framelenp, const ip_pair *const ip_pairp,
	gchar **const ofnamepp, gboolean sep_needed);

static void split_udp(const struct udphdr *const udp_hdrp,
	gint32 *const framelenp, const ip_pair *const ip_pairp,
	gchar **const ofnamepp, gboolean sep_needed);

static void write_dump(gchar *const dump_name, gchar *name_end,
	gboolean sep_needed);

static void cleanup(const gboolean abort);
static void close_all_out(const gboolean exit, const gboolean abort);
static void sighand(int);

int
main(int argc, char *argv[])
{
	int         opt;

	sd.out_file_type = WTAP_FILE_PCAP;

	while ((opt = getopt(argc, argv, "bd:eF:himpt:vw:")) != -1) {
		switch (opt) {

		case 'b':
			sd.split_flags |= BI_DIR;
			break;

		case 'd':
			sd.out_dir = (gchar *)optarg;
			break;

		case 'e':
			sd.split_type |= BY_ENCAP;
			break;

		case 'F':
			sd.out_file_type = wtap_short_string_to_file_type(optarg);
			if (sd.out_file_type < 0) {
				fprintf(stderr, "splitcap: \"%s\" is not a valid capture "
					"file type\n", optarg);
				exit(EXIT_FAILURE);
			}
			break;

		case 'h':
			fprintf(stderr, "splitcap version %s\n", VERSION);
			usage();
			exit(EXIT_SUCCESS);

		case 'i':
			sd.split_type |= BY_IP_ADDR;
			break;

		case 'm':
			sd.split_type |= BY_MAC;
			break;

		case 'p':
			sd.split_type |= BY_PORT;
			break;

		case 't':
			{
				char    *time_end;
				int      err;

				errno = 0;
				sd.time_intvl = strtol(optarg, &time_end, 10);
				if ((err = errno) || sd.time_intvl < 1) {
					err = 1;
					goto tcheck;
				}
				if (*time_end == '\0')
					goto tcheck;
				if (*(time_end + 1) != '\0') {	/* Single character? */
					err = 1;
					goto tcheck;
				}
				switch (tolower(*time_end)) {
				case 's':
					if (sd.time_intvl > TIME_T_MAX)
						err = 1;
					break;
				case 'm':
					if (sd.time_intvl <= TIME_T_MAX / 60)
						sd.time_intvl *= 60;
					else
						err = 1;
					break;
				case 'h':
					if (sd.time_intvl <= TIME_T_MAX / 3600)
						sd.time_intvl *= 3600;
					else
						err = 1;
					break;
				case 'd':
					if (sd.time_intvl <= TIME_T_MAX / 86400)
						sd.time_intvl *= 86400;
					else
						err = 1;
					break;
				default:
					err = 1;
				}
tcheck:
				if (err) {
					fprintf(stderr, "splitcap: invalid time interval %s\n",
						optarg);
					exit(EXIT_FAILURE);
				}
			}
			sd.split_type |= BY_TIME;
			if (sd.time_intvl % 86400 != 0) {
				sd.split_flags |= TIME_HH;
				if (sd.time_intvl % 60 != 0)
					sd.split_flags |= TIME_SS;
			}
			break;

		case 'v':
			sd.split_type |= BY_VLAN;
			break;

		case 'w':
			sd.out_fname_prefix = (gchar *)optarg;
			break;

		case '?':
		case ':':
		default:
			exit(EXIT_FAILURE);

		} /* switch (opt) */
	} /* while (opt = getopt()) */

	if (optind >= argc) {
		fprintf(stderr, "splitcap: no filename specified\n");
		exit(EXIT_FAILURE);
	}
	if (sd.split_type == 0)
		sd.split_type = BY_PORT;
	if (setjmp(sd.jmpenv)) {
		cleanup(TRUE);
		exit(EXIT_FAILURE);
	}
	signal(SIGINT,  sighand);
	signal(SIGTERM, sighand);
	sd.dumpnode_chunk = g_mem_chunk_create(dumpnode, 256, G_ALLOC_AND_FREE);
	sd.out_hash = g_hash_table_new(g_str_hash, g_str_equal);

	for (; optind < argc; optind++) {
		split(argv[optind]);
	}
	cleanup(FALSE);
	exit(EXIT_SUCCESS);

}

static void
usage()
{
	int         i;

	fprintf(stderr, "Usage: splitcap [-behimpv] [-d <dir>] [-w <base filename>]\n");
	fprintf(stderr, "   [-t <time interval>[s|m|h|d]] [-F <capture type>] <infile> [<infile>...]\n\n");
	fprintf(stderr, "   -b (bidirectional) outputs both sides of a conversation\n");
	fprintf(stderr, "   -d <dir> sets output directory\n");
	fprintf(stderr, "   -F <capture type> specifies the capture file type to write:\n");
	for (i = 0; i < WTAP_NUM_FILE_TYPES; i++) {
		if (wtap_dump_can_open(i))
			fprintf(stderr, "      %s - %s\n",
				wtap_file_type_short_string(i), wtap_file_type_string(i));
	}
	fprintf(stderr, "      default is libpcap\n");
	fprintf(stderr, "   -h produces this help listing\n");
	fprintf(stderr, "   -w <base filename> sets filename prefix or default filename\n");
	fprintf(stderr, "\nSplit flags may be specified in any combination:\n\n");
	fprintf(stderr, "   -e split by encapsulation type\n");
	fprintf(stderr, "   -i split by IP address\n");
	fprintf(stderr, "   -m split by MAC address\n");
	fprintf(stderr, "   -p split by IP address and port (default)\n");
	fprintf(stderr, "   -t <time interval>[s|m|h|d] (default: seconds) split by time interval\n");
	fprintf(stderr, "   -v split by VLAN\n");

}

static void
split(const char *const in_fname)
{
#define NAME_PROTO "encapsulation-type-xxxxxx_vlan4095_\
ff:ff:ff:ff:ff:ff-ff:ff:ff:ff:ff:ff_\
ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255#65535-\
ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255#65535_2038-19-01T03:14:07"

	int            rderr;
	long           data_offset;
	gint32         framelen;
	const llc     *llcp;
	gconstpointer  l3hdr;
	enum l3type    l3type;
	gconstpointer  l4hdr;
	guint8         l4type;
	gchar          ofname[sizeof NAME_PROTO];
	ip_pair        ip_pair;
	gchar         *ofnamep;
	gboolean       sep_needed;
	gboolean       is_port;
	wtap          *rd_hdl_l;

	if (! (sd.rd_hdl = wtap_open_offline(in_fname, &rderr, FALSE))) {
		fprintf(stderr, "splitcap: unable to open %s: %s\n",
			in_fname, wtap_strerror(rderr));
		_ABORT;
	}

	sd.in_fname = in_fname;
	sd.in_encap = wtap_file_encap(sd.rd_hdl);
	sd.snaplen = wtap_snapshot_length(sd.rd_hdl);

	switch (sd.in_encap) {

	case WTAP_ENCAP_ETHERNET:
	case WTAP_ENCAP_SLL:
	case WTAP_ENCAP_TOKEN_RING:
	case WTAP_ENCAP_FDDI:
	case WTAP_ENCAP_FDDI_BITSWAPPED:
	case WTAP_ENCAP_PER_PACKET:
	/* iptrace sets per-file encapsulation type after reading first record */
	case WTAP_ENCAP_UNKNOWN:
		break;

	default:
		fprintf(stderr, "splitcap: unable to read \"%s\" encapsulation\n"
			"in %s\n",
			wtap_encap_string(sd.in_encap), in_fname);
		_ABORT;

	} /* switch (sd.in_encap) */

	/* read loop */
	while (wtap_read(sd.rd_hdl, &rderr, &data_offset)) {
		sd.pkthdr = wtap_phdr(sd.rd_hdl);
		sd.pseudo_hdr = wtap_pseudoheader(sd.rd_hdl);
		sd.buf_ptr = wtap_buf_ptr(sd.rd_hdl);
		framelen = sd.pkthdr->caplen;
		l4hdr = l3hdr = llcp = NULL;
		ofname[0] = '\0';
		ofnamep = ofname;
		sep_needed = FALSE;

		if (sd.split_type & BY_ENCAP) {
			strcpy(ofnamep, wtap_encap_short_string(sd.pkthdr->pkt_encap));
			ofnamep += strlen(ofnamep);
			sep_needed = TRUE;
		}

		switch (sd.pkthdr->pkt_encap) {

		case WTAP_ENCAP_ETHERNET:
			split_eth((eth_hdr *)sd.buf_ptr, &llcp, &l3hdr, &l3type,
				&framelen, &ofnamep, sep_needed);
			break;

		case WTAP_ENCAP_SLL:
			split_sll((struct sll_header *)sd.buf_ptr, &llcp, &l3hdr, &l3type,
				&framelen, &ofnamep, sep_needed);
			break;

		case WTAP_ENCAP_TOKEN_RING:
			split_tr((tr_hdr *)sd.buf_ptr, &llcp,
				&framelen, &ofnamep, sep_needed);
			break;

		case WTAP_ENCAP_FDDI:
			split_fddi((fddi_hdr *)sd.buf_ptr, &llcp, &framelen,
				&ofnamep, sep_needed, FALSE);
			break;

		case WTAP_ENCAP_FDDI_BITSWAPPED:
			split_fddi((fddi_hdr *)sd.buf_ptr, &llcp, &framelen,
				&ofnamep, sep_needed, TRUE);
			break;

		default:
			;

		} /* switch (sd.pkthdr->pkt_encap) */

		if (ofnamep > ofname)
			sep_needed = TRUE;

		if (llcp)
			split_llc(llcp, &l3hdr, &l3type, &framelen, &ofnamep, sep_needed);

		if (ofnamep > ofname)
			sep_needed = TRUE;

		if (l3hdr) {

			switch (l3type) {

			case l3type_IP_v4:
				split_ip_v4((e_ip *)l3hdr, &l4hdr, &l4type, &framelen,
					&ip_pair, &ofnamep, sep_needed);
				break;

			case l3type_IP_v6:
				split_ip_v6((struct ip6_hdr *)l3hdr, &l4hdr, &l4type, &framelen,
					&ip_pair, &ofnamep, sep_needed);
				break;

			default:
				;

			} /* switch (l3type) */

		} /* if (l3hdr) */

		if (ofnamep > ofname)
			sep_needed = TRUE;

		if (l3hdr && (l3type == l3type_IP_v4 || l3type == l3type_IP_v6)) {
			if (l4hdr && (l4type == IP_PROTO_TCP || l4type == IP_PROTO_UDP)) {
				is_port = TRUE;
				if (l4type == IP_PROTO_TCP)
					split_tcp((struct tcphdr*)l4hdr, &framelen, &ip_pair,
						&ofnamep, sep_needed);
				else
					split_udp((struct udphdr*)l4hdr, &framelen, &ip_pair,
						&ofnamep, sep_needed);
			} else 
				is_port = FALSE;
			if ((sd.split_type & BY_IP_ADDR) &&
				!((sd.split_type & BY_PORT) && is_port)) {
				if (sep_needed) {
					*ofnamep++ = FIELD_SEP;
				}
				if ((sd.split_flags & BI_DIR) &&
					(ip_pair.flags & IP_PAIR_SRC_GT)) {
					strcpy(ofnamep, ip_pair.dst_addr);
					ofnamep += strlen(ip_pair.dst_addr);
					*ofnamep++ = ADDR_SEP;
					strcpy(ofnamep, ip_pair.src_addr);
					ofnamep += strlen(ip_pair.src_addr);
				} else {
					strcpy(ofnamep, ip_pair.src_addr);
					ofnamep += strlen(ip_pair.src_addr);
					*ofnamep++ = ADDR_SEP;
					strcpy(ofnamep, ip_pair.dst_addr);
					ofnamep += strlen(ip_pair.dst_addr);
				}
			}
		}

		g_assert(framelen >= 0);
		if (ofnamep > ofname)
			sep_needed = TRUE;
		write_dump(ofname, ofnamep, sep_needed);

	} /* read loop */

	if (rderr != 0) {
		fprintf(stderr, "splitcap: error reading %s: %s\n",
			in_fname, wtap_strerror(rderr));
		_ABORT;
	}

	rd_hdl_l = sd.rd_hdl;
	sd.rd_hdl = NULL;
	wtap_close(rd_hdl_l);

}

static void
split_eth(const eth_hdr *const eth_hdrp, const llc **const llcp,
	gconstpointer *const l3hdrp, enum l3type *const l3type,
	gint32 *const framelenp,
	gchar **const ofnamepp, gboolean sep_needed)
{
	gint32           eth_hdr_sz = sizeof(eth_hdr);
	guint16          eth_type;
	ether_vlan      *vlanp;
	const gchar     *vlan_str = "vlan";

	if (*framelenp < (gint32)sizeof(eth_hdr))
		return;
	*framelenp -= sizeof(eth_hdr);

	if (g_ntohs(eth_hdrp->type) == ETHERTYPE_VLAN) {
		if (*framelenp < (gint32)sizeof(ether_vlan))
			return;
		*framelenp -= sizeof(ether_vlan);
		eth_hdr_sz += sizeof(ether_vlan);
		vlanp = (ether_vlan*)((gpointer)eth_hdrp + sizeof(eth_hdr));
		if (sd.split_type & BY_VLAN) {
			if (sep_needed) {
				**ofnamepp = FIELD_SEP;
				(*ofnamepp)++;
			}
			sep_needed = TRUE;
			strcpy(*ofnamepp, vlan_str);
			*ofnamepp += strlen(vlan_str);
			sprintf(*ofnamepp, "%d", g_ntohs(vlanp->tci) & 0xfff);
			*ofnamepp += strlen(*ofnamepp);
		}
		eth_type = g_ntohs(vlanp->type);
	} else {
		eth_type = g_ntohs(eth_hdrp->type);
	}

	if (sd.split_type & BY_MAC) {
		if (sep_needed) {
			**ofnamepp = FIELD_SEP;
			(*ofnamepp)++;
		}
		sep_needed = TRUE;
		if ((sd.split_flags & BI_DIR) &&
			memcmp(&eth_hdrp->src, &eth_hdrp->dst, sizeof eth_hdrp->src) > 0) {
			strcpy(*ofnamepp, ether_to_str(eth_hdrp->dst));
			*(*ofnamepp+ETHER_STRLEN) = ADDR_SEP;
			*ofnamepp += ETHER_STRLEN + 1;		
			strcpy(*ofnamepp, ether_to_str(eth_hdrp->src));
			*ofnamepp += ETHER_STRLEN;
		} else {
			strcpy(*ofnamepp, ether_to_str(eth_hdrp->src));
			*(*ofnamepp+ETHER_STRLEN) = ADDR_SEP;
			*ofnamepp += ETHER_STRLEN + 1;		
			strcpy(*ofnamepp, ether_to_str(eth_hdrp->dst));
			*ofnamepp += ETHER_STRLEN;
		}
	}

	if (eth_type <= IEEE_802_3_MAX_LEN && eth_type != ETHERTYPE_UNK) {
		if (*framelenp >= 2 && *((guint8 *)eth_hdrp + eth_hdr_sz) == 0xff &&
			*((guint8 *)eth_hdrp + eth_hdr_sz + 1) == 0xff) {
			/* Raw 802.3 for IPX/SPX -- not implemented */
		}
		else
			*llcp = (llc *)((guint8 *)eth_hdrp + eth_hdr_sz);
	} else if (eth_type >= ETHERTYPE_MIN || eth_type == ETHERTYPE_UNK) {
		*l3hdrp = (guint8 *)eth_hdrp + eth_hdr_sz;
		*l3type = decode_ether_type(eth_type);
	}
	/* Else protocol error */

}

static void
split_sll(const struct sll_header *const sll_hdrp, const llc **const llcp,
	gconstpointer *const l3hdrp, enum l3type *const l3type,
	gint32 *const framelenp,
	gchar **const ofnamepp _U_, gboolean sep_needed _U_)
{
	guint16          protocol;

	if (*framelenp < SLL_HDR_LEN)
		return;
	*framelenp -= SLL_HDR_LEN;

	protocol = g_ntohs(sll_hdrp->sll_protocol);

	if (protocol < ETHERTYPE_MIN) {
		switch(protocol) {

		case LINUX_SLL_P_802_2:
			*llcp = (llc *)((guint8 *)sll_hdrp + SLL_HDR_LEN);
			break;

		case LINUX_SLL_P_802_3:
		default:
			;

		}
	} else {
		*l3type = decode_ether_type(protocol);
		*l3hdrp = ((gpointer)sll_hdrp) + SLL_HDR_LEN;
	}

}

static void
split_tr(const tr_hdr *const tr_hdrp, const llc **const llcp,
	gint32 *const framelenp,
	gchar **const ofnamepp, gboolean sep_needed)
{
	int              frame_type;
	int              rif_len = 0;
	guint8           tr_src[6];

	if (*framelenp < (gint32)sizeof(tr_hdr))
		return;
	*framelenp -= sizeof(tr_hdr);

	if (sd.split_type & BY_MAC) {
		if (sep_needed) {
			**ofnamepp = FIELD_SEP;
			(*ofnamepp)++;
		}
		memcpy(&tr_src, &tr_hdrp->src, sizeof tr_src);
		/* Clear RII */
		tr_src[0] &= 0x7f;
		sep_needed = TRUE;
		if ((sd.split_flags & BI_DIR) &&
			memcmp(tr_src, &tr_hdrp->dst, sizeof tr_src) > 0) {
			strcpy(*ofnamepp, ether_to_str(tr_hdrp->dst));
			*(*ofnamepp+ETHER_STRLEN) = ADDR_SEP;
			*ofnamepp += ETHER_STRLEN + 1;		
			strcpy(*ofnamepp, ether_to_str(tr_src));
			*ofnamepp += ETHER_STRLEN;
		} else {
			strcpy(*ofnamepp, ether_to_str(tr_src));
			*(*ofnamepp+ETHER_STRLEN) = ADDR_SEP;
			*ofnamepp += ETHER_STRLEN + 1;		
			strcpy(*ofnamepp, ether_to_str(tr_hdrp->dst));
			*ofnamepp += ETHER_STRLEN;
		}
	}

	if (tr_hdrp->src[0] & 0x80) {
		/* Source routed.  Must be at least the two-byte
		   routing control field in the RIF */
		if (*framelenp < 2)
			return;
		rif_len = *((guint8 *)tr_hdrp + sizeof(tr_hdr)) & 0x1f;
		if (*framelenp < rif_len)
			return;
		*framelenp -= rif_len;
	}

	frame_type = (tr_hdrp->fc & 0xc0) >> 6;
	if (frame_type == 1)
		*llcp = (gpointer)tr_hdrp + sizeof(tr_hdr) + rif_len;

}

static void
swap_mac_addr(guint8 *swapped_addr, const guint8 *orig_addr)
{
	int i;

	for (i = 0; i < 6; i++) {
		swapped_addr[i] = BIT_SWAP(orig_addr[i]);
	}
}

static void
split_fddi(const fddi_hdr *const fddi_hdrp, const llc * *const llcp,
	gint32 *const framelenp,
	gchar **const ofnamepp, gboolean sep_needed, gboolean bit_swapped)
{

/* See packet-fddi.c and http://www.protocols.com/pbook/lan.htm#FDDI.
 * I'd like to clarify these with ISO 9314-2:1989, but it costs...
 */

#define FDDI_FC_CLFF        0xF0        /* Class/Length/Format bits */
#define FDDI_FC_LLC_ASYNC   0x50        /* Async. LLC frame */
#define FDDI_FC_ASYNC_R     0x08        /* Reserved */

	int              rif_len = 0;
	guint8           fddi_dst[6];
	guint8           fddi_src[6];

	if (*framelenp < (gint32)sizeof(fddi_hdr))
		return;
	*framelenp -= sizeof(fddi_hdr);

	if (bit_swapped) {
		swap_mac_addr(fddi_dst, fddi_hdrp->dst);
		swap_mac_addr(fddi_src, fddi_hdrp->src);
	} else {
		memcpy(&fddi_dst, &fddi_hdrp->dst, sizeof fddi_dst);
		memcpy(&fddi_src, &fddi_hdrp->src, sizeof fddi_src);
	}

	if (sd.split_type & BY_MAC) {
		if (sep_needed) {
			**ofnamepp = FIELD_SEP;
			(*ofnamepp)++;
		}
		/* Clear RII */
		fddi_src[0] &= 0xbf;
		sep_needed = TRUE;
		if ((sd.split_flags & BI_DIR) &&
			memcmp(fddi_src, fddi_dst, sizeof fddi_src) > 0) {
			strcpy(*ofnamepp, ether_to_str(fddi_dst));
			*(*ofnamepp+ETHER_STRLEN) = ADDR_SEP;
			*ofnamepp += ETHER_STRLEN + 1;		
			strcpy(*ofnamepp, ether_to_str(fddi_src));
			*ofnamepp += ETHER_STRLEN;
		} else {
			strcpy(*ofnamepp, ether_to_str(fddi_src));
			*(*ofnamepp+ETHER_STRLEN) = ADDR_SEP;
			*ofnamepp += ETHER_STRLEN + 1;
			strcpy(*ofnamepp, ether_to_str(fddi_dst));
			*ofnamepp += ETHER_STRLEN;
		}
	}

	if (fddi_src[0] & 0x40) {
		/* Source routed.  Must be at least the two-byte
		   routing control field in the RIF */
		if (*framelenp < 2)
			return;
		rif_len = *((guint8 *)fddi_hdrp + sizeof(fddi_hdr)) & 0x1f;
		if (*framelenp < rif_len)
			return;
		*framelenp -= rif_len;
	}

	switch (fddi_hdrp->fc & FDDI_FC_CLFF) {

	case FDDI_FC_LLC_ASYNC:
		if (! (fddi_hdrp->fc & FDDI_FC_ASYNC_R))
			*llcp = (gpointer)fddi_hdrp + sizeof(fddi_hdr) + rif_len;
		break;

	default:
		;

	}

}

static void
split_llc(const llc *const llcp, gconstpointer *const l3hdrp,
	enum l3type *const l3type, gint32 *const framelenp,
	gchar **const ofnamepp _U_, gboolean sep_needed _U_)
{
/* Mask out DSAP I/G bit and SSAP C/R bit */
#define SAP_MASK 0xFE

	snap            *snapp;
	guint32          oui;
	guint16          type;
	guint8           dsap;
	guint8           ssap;

	if (*framelenp < (gint32)sizeof(llc))
		return;
	*framelenp -= sizeof(llc);

	dsap = (llcp->dsap & SAP_MASK);
	ssap = (llcp->ssap & SAP_MASK);

	/* UI command? */
	if (llcp->control != 0x03)
		return;

	if (dsap == SAP_SNAP && ssap == SAP_SNAP) {
		if (*framelenp < (gint32)sizeof(snap))
			return;
		*framelenp -= sizeof(snap);
		snapp = (snap *)((guint8 *)llcp + sizeof(llc));
		oui = snapp->oui[0] << 16 | snapp->oui[1] << 8 | snapp->oui[2];
		type = snapp->type[0] << 8 | snapp->type[1];

		switch (oui) {

		case OUI_ENCAP_ETHER:
			*l3type = decode_ether_type(type);
			*l3hdrp = (guint8 *)snapp + sizeof(snap);
			break;
		
		default:
			;

		}
	}
}

static enum l3type
decode_ether_type(guint16 eth_type)
{
	switch (eth_type) {

	case ETHERTYPE_IP:
		return(l3type_IP_v4);

	case ETHERTYPE_IPv6:
		return(l3type_IP_v6);

	default:
		return(l3type_notinterested);

	}

}

static void
split_ip_v4(const e_ip *const ip_hdrp, gconstpointer *const l4hdrp,
	guint8 *l4type, gint32 *const framelenp, ip_pair *const ip_pairp,
	gchar **const ofnamepp _U_, gboolean sep_needed _U_)
{
#define IP_OFFSET	0x1FFF		/* "Fragment Offset" part	*/

	gint32           hdr_len;
	int              comp;

	if (((ip_hdrp->ip_v_hl>>4) != 4) || *framelenp < (gint32)sizeof(e_ip))
		return;
	hdr_len = (ip_hdrp->ip_v_hl & 0xf) * 4;
	if (*framelenp < hdr_len)
		return;
	*framelenp -= hdr_len;

	if (sd.split_type & NEED_IP_PAIR) {
		/* ip_to_str_buf() takes network byte order */
		ip_to_str_buf((guint8 *)&ip_hdrp->ip_src, ip_pairp->src_addr);
		ip_to_str_buf((guint8 *)&ip_hdrp->ip_dst, ip_pairp->dst_addr);
		/* Compare IP addresses in network byte order */
		comp = memcmp(&ip_hdrp->ip_src, &ip_hdrp->ip_dst,
			sizeof ip_hdrp->ip_src);
		if (comp > 0)
			ip_pairp->flags = IP_PAIR_SRC_GT;
		else if (comp < 0)
			ip_pairp->flags = 0;
		else ip_pairp->flags = IP_PAIR_EQ;
	}

	/* If the packet's fragmented, we can still return the
     * layer 4 header if the fragment offset is 0.
	 */
	if (! (g_ntohs(ip_hdrp->ip_off) & IP_OFFSET)) {
		*l4hdrp = (gpointer)ip_hdrp + hdr_len;
		*l4type = ip_hdrp->ip_p;
	}

}

static void
split_ip_v6(const struct ip6_hdr *const ip_hdrp, gconstpointer *const l4hdrp,
	guint8 *l4type, gint32 *const framelenp, ip_pair *const ip_pairp,
	gchar **const ofnamepp _U_, gboolean sep_needed _U_)
{
	int              comp;
	gpointer         hdrp;
	guint8           thistype;
	gint32           hdr_len;
	guint16          frag_off = 0;
	gboolean         more_v6hdrs;

	if (((ip_hdrp->ip6_vfc>>4) != 6) ||
		*framelenp < (gint32)sizeof(struct ip6_hdr))
		return;
	*framelenp -= sizeof(struct ip6_hdr);

	if (sd.split_type & (BY_IP_ADDR | NEED_IP_PAIR)) {
		strcpy(ip_pairp->src_addr, ip6_to_str(&ip_hdrp->ip6_src));
		strcpy(ip_pairp->dst_addr, ip6_to_str(&ip_hdrp->ip6_dst));
		comp = memcmp(&ip_hdrp->ip6_src, &ip_hdrp->ip6_dst,
			sizeof(struct e_in6_addr));
		if (comp > 0)
			ip_pairp->flags = IP_PAIR_SRC_GT;
		else if (comp < 0)
			ip_pairp->flags = 0;
		else ip_pairp->flags = IP_PAIR_EQ;
	}

	thistype = ip_hdrp->ip6_nxt;
	hdrp = (gpointer)ip_hdrp + sizeof(struct ip6_hdr);
	more_v6hdrs = TRUE;
	while (more_v6hdrs) {
		switch (thistype) {

		case IP_PROTO_HOPOPTS:
		case IP_PROTO_ROUTING:
		case IP_PROTO_DSTOPTS:
			if (*framelenp < (gint32)sizeof(struct ip6_hbh))
				return;
			hdr_len = ((struct ip6_hbh *)hdrp)->ip6h_len * 8 + 1;
			if (*framelenp < hdr_len)
				return;
			*framelenp -= hdr_len;
			thistype = ((struct ip6_hbh *)hdrp)->ip6h_nxt;
			hdrp += hdr_len;
			break;

		case IP_PROTO_FRAGMENT:
			if (*framelenp < (gint32)sizeof(struct ip6_frag))
				return;
			*framelenp -= sizeof(struct ip6_frag);
			frag_off = g_ntohs(((struct ip6_frag *)hdrp)->ip6f_offlg) &
				IP6F_OFF_MASK;
			thistype = ((struct ip6_frag *)hdrp)->ip6f_nxt;
			hdrp += sizeof(struct ip6_frag);
			break;

		case IP_PROTO_NONE:
			return;

		default:
			more_v6hdrs = FALSE;
			break;

		}
	}

	/* If the packet's fragmented, we can still return the
     * layer 4 header if the fragment offset is 0.
	 */
	if (! frag_off) {
		*l4hdrp = hdrp;
		*l4type = thistype;
	}

}

#define ADDR_PORT_PROTO "ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255#65535"

static void
split_tcp(const struct tcphdr *const tcp_hdrp, gint32 *const framelenp,
	const ip_pair *const ip_pairp, gchar **const ofnamepp,
	gboolean sep_needed)
{
	gint32           hdr_len;
	gchar            addr_port_src[sizeof ADDR_PORT_PROTO];
	gchar            addr_port_dst[sizeof ADDR_PORT_PROTO];
	gchar           *ap_p;

	if (*framelenp < (gint32)sizeof(struct tcphdr))
		return;
	hdr_len = (tcp_hdrp->doff) * 4;
	if (*framelenp < hdr_len)
		return;
	*framelenp -= hdr_len;

	if (sd.split_type & BY_PORT) {
		if (sep_needed) {
			**ofnamepp = FIELD_SEP;
			(*ofnamepp)++;
		}
		strcpy(addr_port_src, ip_pairp->src_addr);
		ap_p = addr_port_src + strlen(addr_port_src);
		*ap_p++ = PORT_SEP;
		sprintf(ap_p, "%d", g_ntohs(tcp_hdrp->source));

		strcpy(addr_port_dst, ip_pairp->dst_addr);
		ap_p = addr_port_dst + strlen(addr_port_dst);
		*ap_p++ = PORT_SEP;
		sprintf(ap_p, "%d", g_ntohs(tcp_hdrp->dest));

		if ((sd.split_flags & BI_DIR) && (
			(ip_pairp->flags & IP_PAIR_SRC_GT) ||
			((ip_pairp->flags & IP_PAIR_EQ) &&
				(memcmp(&tcp_hdrp->source, &tcp_hdrp->dest,
					sizeof tcp_hdrp->source) > 0))
			)) {
			strcpy(*ofnamepp, addr_port_dst);
			*ofnamepp += strlen(addr_port_dst);
			**ofnamepp = ADDR_SEP;
			(*ofnamepp)++;
			strcpy(*ofnamepp, addr_port_src);
			*ofnamepp += strlen(addr_port_src);
		} else {
			strcpy(*ofnamepp, addr_port_src);
			*ofnamepp += strlen(addr_port_src);
			**ofnamepp = ADDR_SEP;
			(*ofnamepp)++;
			strcpy(*ofnamepp, addr_port_dst);
			*ofnamepp += strlen(addr_port_dst);
		}
	}

}

static void
split_udp(const struct udphdr *const udp_hdrp, gint32 *const framelenp,
	const ip_pair *const ip_pairp, gchar **const ofnamepp,
	gboolean sep_needed)
{
	gchar            addr_port_src[sizeof ADDR_PORT_PROTO];
	gchar            addr_port_dst[sizeof ADDR_PORT_PROTO];
	gchar           *ap_p;

	if (*framelenp < (gint32)sizeof(struct udphdr))
		return;
	*framelenp -= sizeof(struct udphdr);

	if (sd.split_type & BY_PORT) {
		if (sep_needed) {
			**ofnamepp = FIELD_SEP;
			(*ofnamepp)++;
		}
		strcpy(addr_port_src, ip_pairp->src_addr);
		ap_p = addr_port_src + strlen(addr_port_src);
		*ap_p++ = PORT_SEP;
		sprintf(ap_p, "%d", g_ntohs(udp_hdrp->source));

		strcpy(addr_port_dst, ip_pairp->dst_addr);
		ap_p = addr_port_dst + strlen(addr_port_dst);
		*ap_p++ = PORT_SEP;
		sprintf(ap_p, "%d", g_ntohs(udp_hdrp->dest));

		if ((sd.split_flags & BI_DIR) && (
			(ip_pairp->flags & IP_PAIR_SRC_GT) ||
			((ip_pairp->flags & IP_PAIR_EQ) &&
				(memcmp(&udp_hdrp->source, &udp_hdrp->dest,
					sizeof udp_hdrp->source) > 0))
			)) {
			strcpy(*ofnamepp, addr_port_dst);
			*ofnamepp += strlen(addr_port_dst);
			**ofnamepp = ADDR_SEP;
			(*ofnamepp)++;
			strcpy(*ofnamepp, addr_port_src);
			*ofnamepp += strlen(addr_port_src);
		} else {
			strcpy(*ofnamepp, addr_port_src);
			*ofnamepp += strlen(addr_port_src);
			**ofnamepp = ADDR_SEP;
			(*ofnamepp)++;
			strcpy(*ofnamepp, addr_port_dst);
			*ofnamepp += strlen(addr_port_dst);
		}
	}

}

static void
write_dump(gchar *const dump_name, gchar *name_end, gboolean sep_needed)
{
/* ISO 8601 */
#define CURRENT_TIME_PROTO "1970-01-01T00:00:00"
#define STRFDATE "%Y-%m-%d"
#define STRFHHMM "T%H:%M"
#define STRFSS ":%S"

	GString         *newpath = NULL;
	dumpnode        *dumpp;
	struct stat		 statb;
	int              wrerr;
	struct tm       *tm;
	char             strf[strlen(STRFDATE) + strlen(STRFHHMM) +
	                    strlen(STRFSS) + 1];
	char            *strfp = strf;

	static time_t    time_bound_low;
	static time_t    time_bound_high;
	static char      current_time[sizeof CURRENT_TIME_PROTO];
	static gchar     *nullstr = "";

	if (sd.split_type & BY_TIME) {
		if (sep_needed)
			*name_end++ = FIELD_SEP;
		sep_needed = TRUE;
		if (sd.pkthdr->ts.tv_sec <  time_bound_low ||
		    sd.pkthdr->ts.tv_sec >= time_bound_high) {
			/* Not first time through, existing output files? */
			if (time_bound_low < time_bound_high)
				/* Assume input is in time order, and we
				 * don't need the existing output files.
				 */
				close_all_out(FALSE, FALSE);
			/* Cache the representation for efficiency */
			time_bound_low = sd.pkthdr->ts.tv_sec / sd.time_intvl *
				sd.time_intvl;
			time_bound_high = time_bound_low + sd.time_intvl;
			tm = localtime(&time_bound_low);
			strcpy(strfp, STRFDATE);
			if (sd.split_flags & TIME_HH) {
				strfp += strlen(STRFDATE);
				strcpy(strfp, STRFHHMM);
				if (sd.split_flags & TIME_SS) {
					strfp += strlen(STRFHHMM);
					strcpy(strfp, STRFSS);
				}
			}
			strftime(current_time, sizeof current_time, (const char *)&strf,
				tm);
		}
		strcpy(name_end, current_time);
	}

	if (dump_name[0] == '\0') {
		/* Nothing interesting in packet, so write to default file */
		if (sd.dflt_dumpnp == NULL) {
			/* Default file not open yet */
			newpath = g_string_sized_new(256);
			if (sd.out_fname_prefix == NULL) {
				/* key used only for filename for error messages */
				sd.dflt_dumpnode.key = DFLT_PREFIX;
				newpath = g_string_assign(newpath, DFLT_PREFIX);
			} else {
				/* -w option */
				sd.dflt_dumpnode.key = sd.out_fname_prefix;
				newpath = g_string_assign(newpath, sd.out_fname_prefix);
			}
			sd.dflt_dumpnp = &sd.dflt_dumpnode;
		}
		dumpp = sd.dflt_dumpnp;
	} else {
		if (! (dumpp = g_hash_table_lookup(sd.out_hash, dump_name))) {
			/* New output file */
			newpath = g_string_sized_new(256);
			if (sd.out_fname_prefix != NULL) {
				/* -w option */
				newpath = g_string_assign(newpath, sd.out_fname_prefix);
				newpath = g_string_append_c(newpath, FIELD_SEP);
			}
			newpath = g_string_append(newpath, dump_name);
			dumpp = g_chunk_new0(dumpnode, sd.dumpnode_chunk);
			/* Unfortunately, the glist_* functions don't understand list
			 * tails.  For instance, the g_list_append() function chugs
			 * through the list to find the end.  Therefore we, slightly
			 * awkwardly, grow the list of open files by prepending to the
			 * head.  The most recently opened file is on the left and the
			 * least recently opened file is on the right.
             */
			sd.out_open_queue_head = g_list_prepend(
				sd.out_open_queue_head, dumpp);
			if (sd.out_open_queue_tail == NULL)
				sd.out_open_queue_tail = sd.out_open_queue_head;
			dumpp->key = g_strdup(dump_name);
			dumpp->open_queue_entry = sd.out_open_queue_head;
			g_hash_table_insert(sd.out_hash, dumpp->key, dumpp);
		}
	}

	if (newpath == NULL) {
		/* Assume no dump routine can write mixed encapsulation types */
		if (sd.pkthdr->pkt_encap != dumpp->filetype) {
			fprintf(stderr, "splitcap: unable to write \"%s\" encapsulated "
				"record to\n   %s%s,\n   opened for \"%s\" encapsulation: "
				"try using -e flag\n",
				wtap_encap_string(sd.pkthdr->pkt_encap),
				dumpp->key,
				dumpp->filename_suffix,
				wtap_encap_string(dumpp->filetype));
			_ABORT;
		}
	} else {
		dumpp->filetype = sd.pkthdr->pkt_encap;
		if (sd.out_dir != NULL) {
			/* -d option */
			newpath = g_string_prepend_c(newpath, G_DIR_SEPARATOR);
			newpath = g_string_prepend(newpath, sd.out_dir);
		}
		if (!stat(newpath->str, &statb)) {
			/* Opening an exising output file will clobber it,
			 * so construct a unique filename.
			 */
			guint        n;
			gint         baselen;
			const gchar  unique_proto[] = "~9999";
			gchar        number[4+1];
			gint         numlen;
			gboolean     unique_found;

			/* Make sure it's long enough */
			newpath = g_string_append(newpath, (const gchar *)&unique_proto);
			/* Length with trailing ~ */
			baselen = newpath->len - (sizeof unique_proto - 1 - 1);
			for (n=1; n < (int)1E4; n++) {
				sprintf((char *)&number, "%d", n);
				numlen = strlen((char *)&number);
				memcpy(newpath->str + baselen, &number, numlen + 1);
				newpath->len = baselen + numlen;
				if ((unique_found = (gboolean)stat(newpath->str, &statb)))
					break;
			}
			if (unique_found)
				dumpp->filename_suffix = g_strdup(newpath->str + baselen - 1);
			else {
				fprintf(stderr, "splitcap: unable to make unique filename "
					"for\n   %s\n", dumpp->key);
				_ABORT;
			}
		} /* Filename clash */
		else
			dumpp->filename_suffix = nullstr;

		/* Loop until we can open a file without hitting the ulimit */
		for (;;) {
			GList           *open_queue_close;
			GList           *open_queue_prev;
			dumpnode        *dump_close;

			if (sd.split_flags & OPEN_MAX_HIT) {
				/* The limit on the number of open files has been reached.
				 * Close the earliest opened file.
				 */
				if (sd.out_open_queue_head == NULL ||
					sd.out_open_queue_head == sd.out_open_queue_tail) {
					/* 0 or 1 non-default files are open.  Bear in mind that
                     * we might have just added a file to the "open" queue
                     * which isn't open, but this makes no difference.
                     */
					fprintf(stderr, "splitcap: can't open output file; "
						"ulimit is too low\n");
					_ABORT;
				}
				open_queue_close = sd.out_open_queue_tail;
				dump_close = (dumpnode*)open_queue_close->data;
				open_queue_prev = g_list_previous(open_queue_close);
				open_queue_prev->next = NULL;
				sd.out_open_queue_tail = open_queue_prev;
				g_list_free_1(open_queue_close);
				g_hash_table_remove(sd.out_hash, dump_close->key);
			    if (! wtap_dump_close(dump_close->wr_hdl, &wrerr)) {
					fprintf(stderr, "splitcap: error closing %s%s:\n   %s\n",
						dump_close->key, dump_close->filename_suffix,
						wtap_strerror(wrerr));
					_ABORT;
				}
				g_free(dump_close->key);
				if (dump_close->filename_suffix[0] != '\0')
					g_free(dump_close->filename_suffix);
				g_chunk_free(dump_close, sd.dumpnode_chunk);
			} /* Limit reached */
			if ((dumpp->wr_hdl = wtap_dump_open(newpath->str, sd.out_file_type,
				sd.pkthdr->pkt_encap, sd.snaplen, &wrerr)))
				break;
			if (wrerr == EMFILE) {
				sd.split_flags |= OPEN_MAX_HIT;
				continue;
			}
			if (wrerr < 0) {	/* Wiretap error: give more details */
				fprintf(stderr, "splitcap: can't create output file \"%s%s\"\n"
					"   with format \"%s\" from input file \"%s\"\n"
					"   with encapsulation (network) type \"%s\":\n   %s\n",
					dumpp->key,
					dumpp->filename_suffix,
					wtap_file_type_short_string(sd.out_file_type),
					sd.in_fname,
					wtap_encap_short_string(sd.pkthdr->pkt_encap),
					wtap_strerror(wrerr));
			} else {
				fprintf(stderr, "splitcap: can't create output file \"%s%s\"\n"
					"   from input file \"%s\":\n   %s\n",
					dumpp->key, dumpp->filename_suffix, sd.in_fname,
					wtap_strerror(wrerr));
			}
			_ABORT;
		} /* Open loop */
		g_string_free(newpath, TRUE);
	} /* New output file */

	if (!wtap_dump(dumpp->wr_hdl, sd.pkthdr,
		sd.pseudo_hdr, sd.buf_ptr, &wrerr)) {
		fprintf(stderr, "splitcap: error writing \"%s%s\":\n   %s\n",
			dumpp->key, dumpp->filename_suffix, wtap_strerror(wrerr));
		_ABORT;
	}
}

static void
cleanup(const gboolean abort) {

	wtap         *rd_hdl_l;

	if ((rd_hdl_l = sd.rd_hdl)) {
		sd.rd_hdl = NULL;
		wtap_close(rd_hdl_l);
	}

	close_all_out(TRUE, abort);

}

static void
close_all_out(const gboolean exit, const gboolean abort)
{

	GList            *open_queue_p;
	dumpnode         *dumpp;
	wtap_dumper      *wr_hdl_l;
	int               wrerr;

	/* If we're exiting, there's no point in releasing
	 * the dynamic structures.
	 */

	if ((dumpp = sd.dflt_dumpnp)) {
		sd.dflt_dumpnp = NULL;
		if (! wtap_dump_close(dumpp->wr_hdl, &wrerr)) {
			if (! abort) {
				fprintf(stderr, "splitcap: error closing \"%s%s\": %s\n",
					dumpp->key, dumpp->filename_suffix,
					wtap_strerror(wrerr));
				_ABORT;
			}
		}
		if (! exit && dumpp->filename_suffix[0] != '\0')
			g_free(dumpp->filename_suffix);
	}

	open_queue_p = sd.out_open_queue_head;
	while (open_queue_p) {
		if ((dumpp = open_queue_p->data)) {
			if ((wr_hdl_l = dumpp->wr_hdl)) {
				dumpp->wr_hdl = NULL;
				if (! wtap_dump_close(wr_hdl_l, &wrerr)) {
					if (! abort) {
						fprintf(stderr, "splitcap: error closing \"%s%s\":\n"
							"   %s\n",
							dumpp->key, dumpp->filename_suffix,
							wtap_strerror(wrerr));
						_ABORT;
					}
				}
			}
			if (! exit) {
				g_free(dumpp->key);
				if (dumpp->filename_suffix[0] != '\0')
					g_free(dumpp->filename_suffix);
			}
			open_queue_p->data = NULL;
		}
		open_queue_p = g_list_next(open_queue_p);
	}

	if (! exit) {
		g_list_free(sd.out_open_queue_head);
		sd.out_open_queue_head = sd.out_open_queue_tail = NULL;
		g_mem_chunk_destroy(sd.dumpnode_chunk);
		g_hash_table_destroy(sd.out_hash);
		sd.dumpnode_chunk = g_mem_chunk_create(dumpnode, 256, G_ALLOC_AND_FREE);
		sd.out_hash = g_hash_table_new(g_str_hash, g_str_equal);
	}

}

static void
sighand(int signum _U_)
{
	_ABORT;
}
